# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "fast-vindex"
copyright = "2025, Gwenaël CAËR"
author = "Gwenaël CAËR"
release = "0.0.1"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx_design",
    "myst_nb",
    "sphinx.ext.autosummary",
]

myst_enable_extensions = ["colon_fence"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# nbsphinx configurations
# nbsphinx_execute = "never"

nb_execution_mode = "off"
nb_execution_excludepatterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
html_static_path = ["_static"]
html_favicon = "_static/favicon.ico"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.

# html_sidebars = {"getting-started": [], "contribute": [], "cite": []}

html_theme_options = {
    "logo": {
        "image_light": "_static/fast_vindex_logo_light.png",
        "image_dark": "_static/fast_vindex_logo_dark.png",
    },
    "header_links_before_dropdown": 4,
}



import dask.array as da
import numpy as np
from dask.array.utils import assert_eq

from fast_vindex import patched_vindex


def test_vindex_1d():
    arr = np.arange(9).reshape((9,))
    x = da.from_array(arr, chunks=(3,))
    indexes = (np.array([1, 2, 3]).reshape(1, 3),)

    with patched_vindex():
        result_fv = x.vindex[indexes]

    result = x.vindex[indexes]
    assert_eq(result_fv, result)


def test_vindex_2d():
    arr = np.arange(9 * 9).reshape((9, 9))
    x = da.from_array(arr, chunks=(3, 3))
    indexes = (
        np.array([1, 2, 3]).reshape(1, 3, 1),
        np.array([4, 5, 6]).reshape(1, 1, 3),
    )

    with patched_vindex():
        result_fv = x.vindex[indexes]

    result = x.vindex[indexes]
    assert_eq(result_fv, result)


def test_vindex_3d():
    arr = np.arange(9 * 9 * 9).reshape((9, 9, 9))
    x = da.from_array(arr, chunks=(3, 3, 3))
    indexes = (
        np.array([1, 2, 3]).reshape(1, 3, 1, 1),
        np.array([4, 5, 6]).reshape(1, 1, 3, 1),
        np.array([1, 2, 3]).reshape(1, 1, 1, 3),
    )

    with patched_vindex():
        result_fv = x.vindex[indexes]

    result = x.vindex[indexes]
    assert_eq(result_fv, result)


def test_vindex_4d():
    arr = np.arange(9 * 9 * 9 * 9).reshape((9, 9, 9, 9))
    x = da.from_array(arr, chunks=(3, 3, 3, 3))
    indexes = (
        np.array([1, 2, 3]).reshape(1, 3, 1, 1, 1),
        np.array([4, 5, 6]).reshape(1, 1, 3, 1, 1),
        np.array([1, 2, 3]).reshape(1, 1, 1, 3, 1),
        np.array([4, 5, 6]).reshape(1, 1, 1, 1, 3),
    )

    with patched_vindex():
        result_fv = x.vindex[indexes]

    result = x.vindex[indexes]
    assert_eq(result_fv, result)


def test_vindex_basic():
    """Official test from dask"""
    x = np.arange(56).reshape((7, 8))
    d = da.from_array(x, chunks=(3, 4))

    # cases where basic and advanced indexing coincide

    with patched_vindex():
        result = d.vindex[0]
    assert_eq(result, x[0])

    with patched_vindex():
        result = d.vindex[0, 1]
    assert_eq(result, x[0, 1])

    # Error
    # with patched_vindex():
    #    result = d.vindex[[0, 1], ::-1]  # slices last
    # assert_eq(result, x[:2, ::-1])


# def test_vindex_nd():
#     x = np.arange(56).reshape((7, 8))
#     d = da.from_array(x, chunks=(3, 4))

#     with patched_vindex():
#         result = d.vindex[[[0, 1], [6, 0]], [[0, 1], [0, 7]]]
#     assert_eq(result, x[[[0, 1], [6, 0]], [[0, 1], [0, 7]]])

#     #with patched_vindex():
#     result = d.vindex[np.arange(7)[:, None], np.arange(8)[None, :]]
#     assert_eq(result, x)

#     #with patched_vindex():
#     result = d.vindex[np.arange(7)[None, :], np.arange(8)[:, None]]
#     assert_eq(result, x.T)



import contextlib

import dask.array as da

from fast_vindex.array.core import _vindex


@contextlib.contextmanager
def patched_vindex():
    original = da.core._vindex
    da.core._vindex = _vindex
    try:
        yield
    finally:
        da.core._vindex = original



import logging
from itertools import product
from pathlib import Path
from typing import Literal, Optional

import dask.array as da
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr


def generate_dask_datacube_from_random(shape, chunks):
    """Générer un dask datacube à partir de valeur aléatoire"""
    return da.random.random(shape, chunks=chunks)


def generate_dask_datacube_from_indices(shape, chunks):
    """Générer un dask datacube avec les valeurs des index"""
    # Création d'un numpy array vide
    narray = np.empty((shape), dtype=object)

    # Remplissage de l'array avec les valeurs de ces indices
    # (Ex: narray[3, 4, 2] = (3, 4, 2)
    for indices in product(*(range(s) for s in shape)):
        narray[indices] = indices

    # Création d'un dask array à partir du numpy array
    darray = da.from_array(narray, chunks=chunks)
    return darray


def generate_dataset(n, shape, buffers):
    """Générer un Dataset de point d'observation (entier et non nombre décimaux)"""
    obj = xr.Dataset()
    obs = np.arange(n)
    obj["obs"] = ("obs", obs)
    for i, (size, buffer) in enumerate(zip(shape, buffers)):
        dim = f"x{i}"
        value = np.random.randint(buffer, size - (buffer + 1), n)
        obj[dim] = ("obs", value)

    return obj


def func_array(values, **kwargs):
    """
    Generates an array of ranges centered around each value with a
    specified buffer.
    """
    return np.array(
        [
            np.arange(value - kwargs["buffer"], value + kwargs["buffer"] + 1)
            for value in values
        ]
    )


def _array(dataarray: xr.DataArray, output_core_dims: str, buffer: dict[str, int]):
    """
    Generates an array of ranges centered around each value with a
    specified buffer.
    """
    return xr.apply_ufunc(
        func_array,
        dataarray,
        input_core_dims=[[]],
        output_core_dims=[[output_core_dims]],
        output_dtypes=[int],
        dask_gufunc_kwargs=dict(output_sizes={output_core_dims: 2 * buffer + 1}),
        dask="allowed",
        kwargs={"buffer": buffer},
    ).astype("int32")


def generate_array(dataset, buffers, output=None):
    """Génère à partir du dataset les array des minicubes:
    * 'dataset' pour obtenir un xarray.dataset avec les array des index
    * 'vindex' pour obtenir un tuple utilisable directement dans vindex dask"""
    array = dataset.copy()
    if output == "dataset":
        for idx, ((k, v), buffer) in enumerate(zip(array.data_vars.items(), buffers)):
            array[f"{k}_array"] = _array(v, f"_{k}", buffer)
        return array
    elif output == "vindex":
        size = len(array.data_vars)
        indices = {}
        for idx, ((k, v), buffer) in enumerate(zip(array.data_vars.items(), buffers)):
            n = len(v)
            new_shape = [1] * size
            new_shape[idx] = 2 * buffer + 1
            new_shape = [n] + new_shape
            new_array = _array(v, f"_{k}", buffer)
            indices[k] = new_array.values.reshape(new_shape)
        indices = tuple(indices.values())
        return indices
    else:
        raise ValueError("Choose an output between : 'dataset' and 'vindex'")


def generate_fancy_indexes(x, n, margin):
    offsets = np.arange(-margin, margin)
    indexes = []
    for axis, size in enumerate(x.shape):
        low, high = margin, size - margin
        base_indices = np.random.randint(low, high, size=(n, 1))
        expanded = base_indices + offsets
        shape = [1] * x.ndim
        shape[axis] = 2 * margin
        expanded = expanded.reshape((n, *shape))
        indexes.append(expanded)
    return tuple(indexes)


def estimate_graph_size(collections):
    from collections.abc import Iterator

    import dask
    from dask.base import collections_to_expr
    from dask.utils import format_bytes
    from distributed.protocol import serialize, to_serialize
    from distributed.protocol.serialize import Serialized
    from distributed.utils import nbytes

    # issue de distributed/client.py/compute()
    collections = tuple(
        (dask.delayed(a) if isinstance(a, (list, set, tuple, dict, Iterator)) else a)
        for a in collections
    )
    variables = [a for a in collections if dask.is_dask_collection(a)]

    kwargs = {}
    optimize_graph = True
    expr = collections_to_expr(variables, optimize_graph, **kwargs)

    # issue de distributed/client.py/_graph_to_futures()
    expr_ser = Serialized(*serialize(to_serialize(expr), on_error="raise"))
    pickled_size = sum(map(nbytes, [expr_ser.header] + expr_ser.frames))
    return pickled_size, format_bytes(pickled_size)


# Définition d'une seed pour que la génération des nombres aléatoires soit reproductible
rs = np.random.RandomState(123)


def arange_center(start, stop, size):
    step = (stop - start) / size
    return np.arange(start + step / 2, stop + step / 2, step)


# Datacube
def set_datacube(shape, chunks, n_var=1):
    dcb = xr.Dataset()

    dcb["time"] = ("time", pd.date_range("2020-01-01", periods=shape["time"], freq="d"))
    dcb["lon"] = ("lon", arange_center(-180, 180, shape["lon"]))
    dcb["lat"] = ("lat", arange_center(-90, 90, shape["lat"]))
    for n in range(n_var):
        dcb[f"data_{n}"] = (
            ["time", "lon", "lat"],
            generate_dask_datacube_from_random(
                tuple(shape.values()), tuple(chunks.values())
            ),
        )
    return dcb


# Dataset
def set_dataset(obs, dim):
    dst = xr.Dataset()
    dst["obs"] = ("obs", np.arange(obs))
    dst["time"] = (
        "obs",
        pd.to_datetime(
            rs.choice(
                pd.date_range("2020-01-01", periods=dim["time"], freq="d"),
                size=obs,
                replace=True,
            )
        ),
    )
    dst["lon"] = ("obs", rs.uniform(-180, 180, size=obs))
    dst["lat"] = ("obs", rs.uniform(-90, 90, size=obs))
    return dst


class Array2D:
    """
    Surcouche Numpy pour représenter la sélection d'une zone dans un array 2D.

    arr = np.arange(81).reshape(9, 9)
    a2d = Array2D(arr)
    a2d[indexes]
    """

    def __init__(self, array2d):
        assert array2d.ndim == 2, "L'array doit être 2D"
        self.array = array2d
        self.dims = array2d.shape

    def __getitem__(self, key):
        selection = np.zeros(self.dims, dtype=bool)
        selection[key] = True

        fig, ax = plt.subplots()
        ax.imshow(np.ones(self.dims), cmap="Greys", alpha=0.1)  # fond gris clair
        ax.imshow(selection, cmap="Blues", alpha=0.6)

        # Ajout de la grille (lignes fines entre les cases)
        ax.set_xticks(np.arange(-0.5, self.dims[1], 1), minor=True)
        ax.set_yticks(np.arange(-0.5, self.dims[0], 1), minor=True)
        ax.grid(which="minor", color="black", linestyle="-", linewidth=0.5)

        # Afficher les ticks principaux avec les indices
        ax.set_xticks(np.arange(self.dims[1]))
        ax.set_yticks(np.arange(self.dims[0]))
        ax.set_xticklabels(np.arange(self.dims[1]))
        ax.set_yticklabels(np.arange(self.dims[0]))

        # ax.invert_yaxis()
        ax.xaxis.set_ticks_position("top")
        ax.xaxis.set_label_position("top")
        plt.show()

    def __repr__(self):
        return self.array.__repr__()


class DaskArray2D:
    """
    arr = da.arange(81).reshape(9, 9).rechunk((3, 3))
    da2d = DaskArray2D(arr)
    da2d[indexes]
    """

    def __init__(self, dask_array):
        if not isinstance(dask_array, da.Array):
            raise TypeError("L'argument doit être un dask.array.Array")
        if dask_array.ndim != 2:
            raise ValueError("DaskArray2D ne supporte que les arrays 2D")

        self.arr = dask_array
        self.dims = dask_array.shape
        self.chunk_sizes = dask_array.chunks  # tuple de tuples (par dimension)

    def __getitem__(self, key):
        selection = np.zeros(self.dims, dtype=bool)
        selection[key] = True

        fig, ax = plt.subplots()
        ax.imshow(np.ones(self.dims), cmap="Greys", alpha=0.1)
        ax.imshow(selection, cmap="Blues", alpha=0.6)

        # Ticks majeurs centrés sur les pixels (indices)
        ax.set_xticks(np.arange(self.dims[1]), minor=False)
        ax.set_yticks(np.arange(self.dims[0]), minor=False)

        # Labels pour ticks
        ax.set_xticklabels(np.arange(self.dims[1]))
        ax.set_yticklabels(np.arange(self.dims[0]))

        # Grille mineure entre pixels
        ax.set_xticks(np.arange(-0.5, self.dims[1], 1), minor=True)
        ax.set_yticks(np.arange(-0.5, self.dims[0], 1), minor=True)
        ax.grid(which="minor", color="black", linestyle="-", linewidth=0.5)

        # Fonction utilitaire pour positions cumulées des chunks
        def cumsum_chunks(chunks):
            positions = [0]
            for size in chunks:
                positions.append(positions[-1] + size)
            return positions

        # Positions lignes chunks (à -0.5 pour coller à la grille)
        x_chunk_lines = cumsum_chunks(self.chunk_sizes[1])
        y_chunk_lines = cumsum_chunks(self.chunk_sizes[0])
        x_chunk_lines_shifted = [x - 0.5 for x in x_chunk_lines]
        y_chunk_lines_shifted = [y - 0.5 for y in y_chunk_lines]

        # Tracer lignes de séparation des chunks avec axvline et axhline
        for x in x_chunk_lines_shifted:
            ax.axvline(x=x, color="red", linewidth=2)
        for y in y_chunk_lines_shifted:
            ax.axhline(y=y, color="red", linewidth=2)

        # ax.invert_yaxis()
        ax.xaxis.set_ticks_position("top")
        ax.xaxis.set_label_position("top")
        plt.show()


def setup_logging(
    app_logger: logging.Logger,
    level: str = "info",
    logfile: Optional[Path] = None,
    fmt: str = "%(message)s",
    log_to: Literal["stream", "file", "both"] = "stream",
) -> bool:
    """
    Configure logging output destination and format.

    :param app_logger: Logger to configure
    :param level: Logging level ('debug', 'info', 'warning', 'error')
    :param logfile: Path to log file (required if log_to includes 'file')
    :param fmt: Log message format
    :param log_to: Where to log: 'stream', 'file', or 'both'
    :return: True if setup is successful
    """

    app_logger.setLevel(logging.DEBUG)
    app_logger.handlers.clear()

    handlers = []

    if log_to in ("file", "both"):
        if not logfile:
            raise ValueError("logfile must be provided when log_to is 'file' or 'both'")
        abs_path = logfile.resolve()
        abs_path.parent.mkdir(parents=True, exist_ok=True)

        file_handler = logging.FileHandler(
            abs_path, mode="a", encoding="utf-8", delay=True
        )
        handlers.append(file_handler)

    if log_to in ("stream", "both"):
        stream_handler = logging.StreamHandler()
        handlers.append(stream_handler)

    log_level = level.lower()
    level_map = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
    }

    if log_level not in level_map:
        app_logger.error(f'Unknown log level "{level}"')
        return False

    for handler in handlers:
        handler.setLevel(level_map[log_level])
        handler.setFormatter(logging.Formatter(fmt=fmt))
        app_logger.addHandler(handler)

    return True



import math
from collections import defaultdict
from functools import reduce
from itertools import product
from numbers import Number
from operator import mul

import numpy as np
from dask._task_spec import List, Task, TaskRef
from dask.array import Array
from dask.array.slicing import replace_ellipsis
from dask.base import tokenize
from dask.highlevelgraph import HighLevelGraph
from dask.utils import cached_cumsum, cached_max


def _vindex(x, *indexes):
    """Point wise indexing with broadcasting.

    >>> x = np.arange(56).reshape((7, 8))
    >>> x
    array([[ 0,  1,  2,  3,  4,  5,  6,  7],
           [ 8,  9, 10, 11, 12, 13, 14, 15],
           [16, 17, 18, 19, 20, 21, 22, 23],
           [24, 25, 26, 27, 28, 29, 30, 31],
           [32, 33, 34, 35, 36, 37, 38, 39],
           [40, 41, 42, 43, 44, 45, 46, 47],
           [48, 49, 50, 51, 52, 53, 54, 55]])

    >>> d = from_array(x, chunks=(3, 4))
    >>> result = _vindex(d, [0, 1, 6, 0], [0, 1, 0, 7])
    >>> result.compute()
    array([ 0,  9, 48,  7])
    """

    indexes = replace_ellipsis(x.ndim, indexes)

    nonfancy_indexes = []
    reduced_indexes = []
    for ind in indexes:
        if isinstance(ind, Number):
            nonfancy_indexes.append(ind)
        elif isinstance(ind, slice):
            nonfancy_indexes.append(ind)
            reduced_indexes.append(slice(None))
        else:
            nonfancy_indexes.append(slice(None))
            reduced_indexes.append(ind)

    nonfancy_indexes = tuple(nonfancy_indexes)
    reduced_indexes = tuple(reduced_indexes)

    x = x[nonfancy_indexes]

    array_indexes = {}
    for i, (ind, size) in enumerate(zip(reduced_indexes, x.shape)):
        if not isinstance(ind, slice):
            ind = np.array(ind, copy=True)
            if ind.dtype.kind == "b":
                raise IndexError("vindex does not support indexing with boolean arrays")
            if ((ind >= size) | (ind < -size)).any():
                raise IndexError(
                    "vindex key has entries out of bounds for "
                    "indexing along axis %s of size %s: %r" % (i, size, ind)
                )
            ind %= size
            array_indexes[i] = ind

    if array_indexes:
        x = _vindex_array(x, array_indexes)

    return x


def _vindex_array(x, dict_indexes):
    """Fancy indexing with only NumPy Arrays."""

    token = tokenize(x, dict_indexes)
    try:
        broadcast_shape = np.broadcast_shapes(
            *(arr.shape for arr in dict_indexes.values())
        )

    except ValueError as e:
        # note: error message exactly matches numpy
        shapes_str = " ".join(str(a.shape) for a in dict_indexes.values())
        raise IndexError(
            "shape mismatch: indexing arrays could not be "
            "broadcast together with shapes " + shapes_str
        ) from e
    npoints = math.prod(broadcast_shape)
    axes = [i for i in range(x.ndim) if i in dict_indexes]

    def _subset_to_indexed_axes(iterable):
        for i, elem in enumerate(iterable):
            if i in axes:
                yield elem

    bounds2 = tuple(
        np.array(cached_cumsum(c, initial_zero=True))
        for c in _subset_to_indexed_axes(x.chunks)
    )
    # axis = _get_axis(tuple(i if i in axes else None for i in range(x.ndim)))
    out_name = "vindex-merge-" + token

    # Now compute indices of each output element within each input block
    # The index is relative to the block, not the array.
    block_idxs = tuple(
        np.searchsorted(b, ind, side="right") - 1
        for b, ind in zip(bounds2, dict_indexes.values())
    )
    starts = (b[i] for i, b in zip(block_idxs, bounds2))
    inblock_idxs = []
    for idx, start in zip(dict_indexes.values(), starts):
        a = idx - start
        if len(a) > 0:
            dtype = np.min_scalar_type(np.max(a, axis=None))
            inblock_idxs.append(a.astype(dtype, copy=False))
        else:
            inblock_idxs.append(a)

    chunks = [c for i, c in enumerate(x.chunks) if i not in axes]

    # determine number of points in one single output block.
    # Use the input chunk size to determine this.
    max_chunk_point_dimensions = reduce(
        mul, map(cached_max, _subset_to_indexed_axes(x.chunks))
    )

    n_chunks, remainder = divmod(npoints, max_chunk_point_dimensions)
    chunks.insert(
        0,
        (
            (max_chunk_point_dimensions,) * n_chunks
            + ((remainder,) if remainder > 0 else ())
            if npoints > 0
            else (0,)
        ),
    )
    chunks = tuple(chunks)

    unis = []
    for block_idxs_dim in block_idxs:
        N = block_idxs_dim.shape[0]
        arr = block_idxs_dim.reshape(N, -1)
        uni = [np.unique(arr[i]) for i in range(N)]
        unis.append(uni)

    def cartesian_product_linewise(line):
        return np.array(tuple(product(*line)))

    # Represents for each observation, which chunk should be opened
    chunk_idxs = [cartesian_product_linewise(row) for row in zip(*unis)]

    # --- Prepare working structures ---
    # Each observation may require opening one or more input chunks.
    # Les parties représente les valeurs d'une observation au sein d'un chunk
    input_blocks = (
        []
    )  # For each part, the source chunk to open (coordinates in the chunk grid)
    obs_ids = []  # Observation ID corresponding to this part
    input_slices = []  # For each part, the slice to extract from the source chunk
    output_slices = []  # For each part, the slice to write in the output chunk
    reshape_defs = (
        []
    )  # For each part, the reshape dimensions to apply to index before extraction.

    # --- Loop over observations ---
    for obs_id, (block_coords, block_indices, candidate_chunks) in enumerate(
        zip(zip(*block_idxs), zip(*inblock_idxs), chunk_idxs)
    ):
        # For each candidate chunk associated with this observation
        for chunk_coords in candidate_chunks:
            chunk_coords = tuple(chunk_coords)

            # Per-chunk, per-dimension temporary storage
            reshape_per_dim = []  # reshape definitions per dimension for this chunk
            input_slices_per_dim = []  # input slices per dimension for this chunk
            output_slices_per_dim = []  # output slices per dimension for this chunk

            obs_ids.append(obs_id)

            # --- Loop over dimensions ---
            for dim, coord in enumerate(chunk_coords):
                # Start with a default reshape vector of 1's for all dimensions
                reshape = [1] * len(chunk_coords)

                # Input indices in the chunk (where block_indices match coord)
                mask = block_coords[dim] == coord
                idx_in_chunk = block_indices[dim][mask]

                # Output indices (location in the output chunk)
                size = block_coords[dim].size
                dtype = np.min_scalar_type(size)
                idx_in_output = np.arange(size, dtype=dtype).reshape(
                    block_coords[dim].shape
                )[mask]

                # Update reshape definition for this dimension
                reshape[dim] = len(idx_in_chunk)
                reshape_per_dim.append(reshape)

                # Build slices (assume step = 1)
                input_slices_per_dim.append(
                    (int(idx_in_chunk[0]), int(idx_in_chunk[-1] + 1), None)
                )
                output_slices_per_dim.append(
                    (int(idx_in_output[0]), int(idx_in_output[-1] + 1), None)
                )

            # Save results for this chunk
            input_blocks.append(tuple(int(c) for c in chunk_coords))
            input_slices.append(tuple(input_slices_per_dim))
            output_slices.append(tuple(output_slices_per_dim))
            reshape_defs.append(tuple(reshape_per_dim))

    # --- Build task names ---
    slice_task_name = "vindex-slice-" + token
    merge_task_name = "vindex-merge-" + token
    dsk = {}

    # Single output chunk coordinates (one output chunk for now)
    out_block_coords = (0,) * len(broadcast_shape)

    # --- Group parts by input chunk ---
    grouped = defaultdict(
        lambda: {
            "in_slices": [],
            "out_slices": [],
            "obs_ids": [],
            "reshape_defs": [],
        }
    )

    for in_blk, in_sl, out_sl, obs_id, resh in zip(
        input_blocks, input_slices, output_slices, obs_ids, reshape_defs
    ):
        grouped[in_blk]["in_slices"].append(in_sl)
        grouped[in_blk]["out_slices"].append(out_sl)
        grouped[in_blk]["obs_ids"].append(obs_id)
        grouped[in_blk]["reshape_defs"].append(resh)

    # --- Create _vindex_slice tasks ---
    merge_slices = defaultdict(list)
    merge_values = defaultdict(list)
    merge_obs_ids = defaultdict(list)
    merge_reshape_defs = defaultdict(list)

    for task_idx, (in_blk, data) in enumerate(grouped.items()):
        task_key = (slice_task_name, task_idx)

        dsk[task_key] = Task(
            task_key,
            _vindex_slice,
            TaskRef((x.name,) + in_blk),  # source chunk
            data["in_slices"],  # slice definitions
            data["reshape_defs"],  # reshape definitions
        )

        merge_slices[out_block_coords].append(data["out_slices"])
        merge_values[out_block_coords].append(TaskRef(task_key))
        merge_obs_ids[out_block_coords].append(data["obs_ids"])
        merge_reshape_defs[out_block_coords].append(data["reshape_defs"])

    # --- Create _vindex_merge tasks ---
    for blk_coords in merge_values.keys():
        task_key = (merge_task_name,) + blk_coords

        dsk[task_key] = Task(
            task_key,
            _vindex_merge,
            broadcast_shape,
            List(merge_slices[blk_coords]),
            List(merge_values[blk_coords]),
            List(merge_obs_ids[blk_coords]),
            List(merge_reshape_defs[blk_coords]),
        )

    array = Array(
        HighLevelGraph.from_collections(out_name, dsk, dependencies=[x]),
        out_name,
        chunks=tuple((i,) for i in broadcast_shape),
        dtype=x.dtype,
        meta=x._meta,
    )

    return array


def _vindex_slice(block, slice_group, reshape_group):
    """
    Extract slices from a block according to slice and reshape definitions.

    Parameters
    ----------
    block : np.ndarray
        The data block to slice.
    slice_group : list[list[tuple[int, int, Optional[int]]]]
        Slice definitions, where each slice is a tuple (start, stop, step).
    reshape_group : list[list[tuple[int, ...]]]
        Reshape definitions corresponding to each slice.

    Returns
    -------
    list[np.ndarray]
        Extracted sub-arrays from the block.
    """
    if len(slice_group) != len(reshape_group):
        raise ValueError(
            f"Mismatched lengths: got {len(slice_group)} slice groups "
            f"but {len(reshape_group)} reshape groups."
        )

    results = []
    for slice_defs, reshape_defs in zip(slice_group, reshape_group):
        index_arrays = [
            np.arange(start, stop, step or 1).reshape(shape)
            for (start, stop, step), shape in zip(slice_defs, reshape_defs)
        ]
        results.append(block[tuple(index_arrays)])

    return results


def _vindex_merge(
    output_shape, slices_groups, values_groups, obs_index_groups, reshape_groups
):
    """
    Merge values from different blocks into a single NumPy array,
    according to provided slices and reshape information.

    Parameters
    ----------
    output_shape : tuple[int]
        Shape of the final merged array.
    slices_groups : list[list[tuple[int, int, int]]]
        List of groups of slice definitions for each observation.
    values_groups : list[list[np.ndarray]]
        List of groups of arrays corresponding to slices.
    obs_index_groups : list[list[int]]
        Indices of the observations in the final output.
    reshape_groups : list[list[tuple[int]]]
        Shapes used to reshape the index arrays.

    Returns
    -------
    np.ndarray
        A NumPy array of shape `output_shape` with merged values.
    """

    # Allocate final result array
    first_dtype = values_groups[0][0].dtype
    merged_array = np.empty(output_shape, dtype=first_dtype)

    # Iterate over groups of slices/values/indices/reshapes
    for slice_group, value_group, obs_group, reshape_group in zip(
        slices_groups, values_groups, obs_index_groups, reshape_groups
    ):
        # Each group corresponds to one chunk
        for slice_defs, value, obs_index, reshape_defs in zip(
            slice_group, value_group, obs_group, reshape_group
        ):
            # Build index arrays for this slice
            index_arrays = [
                np.arange(start, stop, step or 1).reshape(shape)
                for (start, stop, step), shape in zip(slice_defs, reshape_defs)
            ]

            # Place the values in the appropriate position of the merged array
            merged_array[obs_index][tuple(index_arrays)] = value

    return merged_array
